# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.race.race_adapter import DATASET_ID, SUBJECT_MAPPING, SUBSET_LIST, RACEAdapter
from llmuses.benchmarks.race.race_adapter import RACEAdapter as DataAdapterClass
from llmuses.models.model_adapter import MultiChoiceModelAdapter as ModelAdapterClass    # noqa
