# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.gsm8k.gsm8k_adapter import DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.gsm8k.gsm8k_adapter import GSM8KAdapter as DataAdapterClass
from llmuses.models.model_adapter import ChatGenerationModelAdapter as ModelAdapterClass    # noqa
