# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.truthful_qa.truthful_qa_adapter import TruthfulQaAdapter, DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.truthful_qa.truthful_qa_adapter import TruthfulQaAdapter as DataAdapterClass
from llmuses.models.model_adapter import ContinuationLogitsModelAdapter as ModelAdapterClass    # noqa
