# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.ceval.ceval_adapter import CEVALAdapter, DATASET_ID, SUBSET_LIST, SUBJECT_MAPPING
from llmuses.benchmarks.ceval.ceval_adapter import CEVALAdapter as DataAdapterClass
from llmuses.models.model_adapter import MultiChoiceModelAdapter as ModelAdapterClass    # noqa
