# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.arc.arc_adapter import ARCAdapter, DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.arc.arc_adapter import ARCAdapter as DataAdapterClass
from llmuses.models.model_adapter import MultiChoiceModelAdapter as ModelAdapterClass    # noqa
