# Copyright (c) Alibaba, Inc. and its affiliates.


from llmuses.benchmarks.bbh.bbh_adapter import DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.bbh.bbh_adapter import BBHAdapter as DataAdapterClass
from llmuses.models.model_adapter import ChatGenerationModelAdapter as ModelAdapterClass    # noqa
