# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.mmlu.mmlu_adapter import DATASET_ID, SUBJECT_MAPPING, SUBSET_LIST, MMLUAdapter
from llmuses.benchmarks.mmlu.mmlu_adapter import MMLUAdapter as DataAdapterClass
from llmuses.models.model_adapter import MultiChoiceModelAdapter as ModelAdapterClass    # noqa
