# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.hellaswag.hellaswag_adapter import HellaSwagAdapter, DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.hellaswag.hellaswag_adapter import HellaSwagAdapter as DataAdapterClass
from llmuses.models.model_adapter import ContinuationLogitsModelAdapter as ModelAdapterClass    # noqa
