# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.competition_math.competition_math_adapter import CompetitionMathAdapter, DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.competition_math.competition_math_adapter import CompetitionMathAdapter as DataAdapterClass
from llmuses.models.model_adapter import ChatGenerationModelAdapter as ModelAdapterClass    # noqa
