# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.general_qa.general_qa_adapter import DATASET_ID, SUBSET_LIST, GeneralQAAdapter
from llmuses.benchmarks.general_qa.general_qa_adapter import GeneralQAAdapter as DataAdapterClass
from llmuses.models.model_adapter import ChatGenerationModelAdapter as ModelAdapterClass